library hi_sitemap;
/* ***********************************************************
hi_sitemap  Version 0.2.1
サイトマップ作成用　loc, priority, lastmod作成ライブラリ

2014-05-31
hi_tag	http://studyhallweb.info/
*********************************************************** */
/*
 * _Loc　は　loc　への変換用
 * _LastMod は lastmod への変換用
 * どちらも _FileProperties からの使用する
 */
import 'dart:io';
import 'dart:convert';
import 'package:hi_files/hi_files.dart';
// import '../../hi_files/lib/hi_files.dart';

part 'hi_stmap_html.dart';

/**
 * オプションから各パラメータをセットする
 *
 * [method]
 * 	int readIni()    iniファイルを読み込む
 * 	int readRobot()  robots.txtを読み込み除外ディレクトリを記録する
 * 	int set()				 readIni()とreadRobot()を実行する
 */
class _Ini {
	String sTopURL;		// URL of site top page (最後に/あり)
	String sTopDir;		// Top Directory of local PC (最後に\なし)
	String sStMap;		// sitemap.xml のフルパス
	String backFile;	// Full path of sitemap backup
	String logFile;		// Full path of log file
	String diffTime;	// Difference with the standard time (ex.+09:00)
	List<String> sDisallow = new List();	// robots.txt での除外指定  (/***/ と表記)

	int set(List<String> options) {
		// デフォルトのセット
		sStMap = 'sitemap.xml';
		diffTime = '+00:00';
		logFile = 'mkstmap.log';
		backFile = 'sitemap.bak';
		// オプションのセット
		for (String s in options) {
		  List<String> sItem = new List(2);	// 分割用
	    if (s.contains('-sitemap=')) {
		    sItem = s.split('=');
		    sStMap = sItem[1].trim();
	    }
	    if (s.contains('-time=')) {
		    sItem = s.split('=');
		    diffTime = sItem[1].trim();
	    }
	    if (s.contains('-work=')) {
	    	// Work は バックアップファイルとログファイルのディレクトリ (最後に\なし)
		    sItem = s.split('=');
		    backFile = sItem[1].trim() + '\\sitemap.bak';
		    logFile = sItem[1].trim() + '\\mkstmap.log';
	    }
		}
		int i =  _readRobots();
		return i;
	}

/**
 * robots.txt を読み込んで Disallow のディレクトリを sitemap 対象外にするために記録する
 * 対象ディレクトリは sDisallow に記録される (最大10個)
 *
 * [return]
 * 0:no Error, other:Error
 */
  int _readRobots() {
    List<String> sReadFile = new List();
    try {
		  File oFile = new File(sTopDir + '\\robots.txt');
		  sReadFile = oFile.readAsLinesSync();  // この1行で読み込みは終了
		  List<String> sItem = new List(2);			// iniファイルのitem分割用
  	  for (int i = 0; i < sReadFile.length; i++) {
  	  	String s = sReadFile[i];
  	  	if (s.contains('Disallow:')) {
  		    sItem = s.split(':');
  		    sDisallow.add(sItem[1].trim());
  	  	}
  	  }
  	  return 0;
    } on Exception catch(e) {             // Anything else that is an exception
      print('exception: $e');
      return -1;
    } catch(e) {                          // No specified type, handles all
      print('unknown: $e');
      return -2;
    }
	}
}

/**
 *  SiteMap のクラス
 *
 *  sLoc
 *  nPriority
 *  sLastmod
 *  nLinks	被リンク数
 *  nFlag		1:sitemap書き出し済み、2:HTML読み込みリンク解析済
 */
class C_Stmap {
	List<String> sLoc = new List();
	List<num> nPriority = new List();
	List<String> sLastmod = new List();
	List<int> nLinks = new List();	// 被リンク数
	List<int> nFlag = new List();		// 処理済フラグ
}

/**
 *  sitemap.xmlの書き出し
 */
class C_writeMap {
	File oFile;

	// sitemap.xml 書き出し、先にchkで存在確認しておくこと
	int write(_Ini cIni, C_Stmap cMap) {
		int iErr = 0;
		try{
			// backup old mapfile
			String sBakup = cIni.backFile;
			File oBakFile = new File(sBakup);			// backup file
		  File oFile = new File(cIni.sStMap);		// new sitemap file
		  if (oBakFile.existsSync()) {
		  	oBakFile.deleteSync(recursive: false);
		  }
		  if (oFile.existsSync()) {
		  	oFile.copySync(sBakup);
		  }
			oFile.openWrite(mode: FileMode.WRITE);
			// Header
			oFile.writeAsStringSync('<?xml version="1.0" encoding="UTF-8"?>\r\n');
			oFile.writeAsStringSync(
				'<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" ', mode: FileMode.APPEND);
				// 毎回 APPEND を付けないとダメらしい
			oFile.writeAsStringSync(
				'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\r\n', mode: FileMode.APPEND);
			oFile.writeAsStringSync(
				'<!-- mkstmap.dart hi_tag http://studyhallweb.info/ -->\r\n', mode: FileMode.APPEND);

			// Date
			for (int i = 0; i < cMap.sLoc.length; i++) {
				String sloc = cMap.sLoc[i];
				// Disallow Check
				bool bDis = false;	// Disallow対象外か否か
				String sDis = sloc.substring(cIni.sTopURL.length, sloc.length);
						// topURL分を除く (http://topurl/aaa/bbb.html -> aaa/bbb.html)
				if(sDis.indexOf('/') != -1) {
					sDis = sDis.substring(0, sDis.indexOf('/'));
						// 一つ目のディレクトリを向き取り (aaa/bbb.html -> aaa)
					sDis = '/' + sDis + '/';
					if (cIni.sDisallow.indexOf(sDis) != -1) bDis = true;	// 対象外に含まれているとき
				}
				if((bDis == false) && (cMap.nLinks[i] > 0)) {
					// robots.txt の Disallow 対象外で、被リンク数が 1 以上であれば 書き出し　(変更点 020 -> 021)
					String sPoint = '1.0';
					if (cMap.nPriority[i] != 1)
						sPoint = cMap.nPriority[i].toString();
					String sTime = cMap.sLastmod[i] + cIni.diffTime;
					oFile.writeAsStringSync('<url>\r\n', mode: FileMode.APPEND);
					oFile.writeAsStringSync('\t<loc>$sloc</loc>\r\n', mode: FileMode.APPEND);
					oFile.writeAsStringSync('\t<priority>$sPoint</priority>\r\n', mode: FileMode.APPEND);
					oFile.writeAsStringSync('\t<lastmod>$sTime</lastmod>\r\n', mode: FileMode.APPEND);
					oFile.writeAsStringSync('</url>\r\n', mode: FileMode.APPEND);
					cMap.nFlag[i] = 1;	// sitemap書き出し済み
				}
			}
			oFile.writeAsStringSync('</urlset>\r\n', mode: FileMode.APPEND);
		} on Exception catch(e) {
			iErr = -1;
		} catch(e) {
			iErr = -2;
			print(e);
		} finally {
			return iErr;
		}
	}
}

/**
 * loc変換用クラス　C_FilePropertysから使用される
 */
class _Loc {
	String sLoc;
	String sTopUrl;		// URL of Top page (最後は'/')
	String sTopDir;		// Dir of Top page (最後に'\'なし)

	// Top の URL, Directory を指定する
	_Loc(String sU, String sF) {
		sTopUrl = sU.substring(0, sU.length - 1);
		sTopDir = sF;
	}

	// ファイル一覧から作成したディレクトリ名とファイル名からURLを作成する
	String getLoc(String sDir, String sFile) {
		// Directory
		if(sTopDir == sDir)
			// ルート直下のとき
			sDir = null;
		else
			sDir = sDir.substring(sTopDir.length + 1, sDir.length);
													// + 1 はディレクトリの \ を消すため
		// File
		if(sFile == 'index.html') {
			if (sDir == null)		// ルート直下の場合
				sLoc = sTopUrl + '/';
			else
				sLoc = sTopUrl + '/' + sDir + '/';
		} else {
			if (sDir == null)
				sLoc = sTopUrl + '/' + sFile;
			else
				sLoc = sTopUrl + '/' + sDir + '/' + sFile;
		}
		return sLoc;
	}
}

/**
 * lastmod 変換用 C_FilePropertiesから使用される
 */
class _LastMod {
	String sDateTime;

	/** Fileオブジェクトのタイムスタンプからsitemap用のlastmodを作成する
	* in sTiemstamp ファイルのタイムスタンプ(toString()で型変換)
	*	Out	lastmod文字列(標準時間との時差はなし)
	*/
	String chg(String sTimestamp) {
		sDateTime = sTimestamp.replaceFirst(' ', 'T');
		sDateTime = sDateTime.substring(0, sDateTime.length - 4);
		return sDateTime;
	}
}

// FileProperties ファイル情報の受け渡しを容易にするため
class _FileProperties {
	String sTopURL;
	String sTopDir;
	String sDir;
	String sFile;
	String sExt;
	String sURL;
	String sTime;	// Timestamp
	String sMod;	// lastmod
	num		nPoint;
	num		nPoint1 = 0;	// 基本点1
	num		nPoint2 = 0;	// 基本点2
	num		nPoint3 = 0;	// タイムスタンプによる減点

	// ローカルファイル名からURLを作成、基本点1も計算
	setLoc() {
		// URL
		_Loc cLoc = new _Loc(sTopURL, sTopDir);
		sURL = cLoc.getLoc(sDir, sFile);

		// 基本点1の計算
		if (sFile == 'index.html') {
			if (sTopDir == sDir)
				nPoint1 = 1;
			else
				nPoint1 = 0.7;
		} else
			nPoint1 = 0.4;
	}

	// タイムスタンプから lastmodを作成、タイムスタンプ減点も計算
	setMod() {
		_LastMod cMod = new _LastMod();
		sMod = cMod.chg(sTime);

		// 減点の計算
		// ★ 日付計算が面倒なので後で作成
	}

	setAll() {
		setLoc();
		setMod();
		if (nPoint1 < nPoint2)
			nPoint = nPoint2 + nPoint3;
		else
			nPoint = nPoint1 + nPoint3;
	}
}

/**
 * make sitemap.xml form local html files
 * and check files that are not linked or disabled links
 *
 * Constructor
 *   C_mkstmap(String top URL of homepage,
 *      String top directory of local files, List<String> options)
 *   option
 *    -sitemap=full path of sitemap.xml
 *    -time=difference with the standard time
 *    -work=workdirectory
 *
 * Metohd
 *   int make()
 *      return error code. 0 is no error.
 */
class C_mkstmap {
	_Ini cIni = new _Ini();

	C_mkstmap(String topURL, String topDir, List<String> options) {
		cIni.sTopURL = topURL;
		cIni.sTopDir = topDir;
		cIni.set(options);
	}

	int make() {
		int iErr = 0;	// Error Nomber
		C_log log = new C_log(cIni.logFile);		// log

		// ローカルのhtmlファイル一覧を取得
		C_FileList cDir = new C_FileList();
		cDir.setPath(cIni.sTopDir, ['html','xml']);
		// C_Dir cDir = new C_Dir(cIni.sTopDir,['html','xml']);
		cDir.getSync();
		print('Local File Coount:' + cDir.filesCount.toString());

		// sitemap格納クラス
		C_Stmap cStmap = new C_Stmap();

		// 子ディレクトリを含む全てのファイルリストを作成
		_FileProperties cFile = new _FileProperties();		// Fiile情報
		cFile.sTopURL = cIni.sTopURL;
		cFile.sTopDir = cIni.sTopDir;
		int i;
		for (i = 0; i < cDir.filesCount; i++ ) {
			cFile
				..sDir = cDir.directory[i]
				..sFile = cDir.filename[i]
				..sTime = cDir.timeStamp[i].toString()
				..setAll();

			// _debugprint(i, cDir.filename[i], cFile.sURL);

			cStmap
				..sLoc.add(cFile.sURL)
				..nPriority.add(cFile.nPoint)
				..sLastmod.add(cFile.sMod)
				..nLinks.add(0)
				..nFlag.add(0);
		}

		// sitemapの書き出し (元の位置)

		// ファイルの数だけ各ファイルを処理
		// この中でリンス数のカウントもしている
		_ReadHTML cHTML = new _ReadHTML();
		cHTML
			..sTopDir = cIni.sTopDir
			..sTopURL = cIni.sTopURL;
		for (i = 0; i < cStmap.nFlag.length; i++ ){
			// HTMLファイル中の aタグからリンクを取得、被リンク数をカウント　(変更点 020 -> 021)
			cHTML
				..sFile = cDir.filename[i]
				..sDir = cDir.directory[i];
			iErr = cHTML.read(cStmap, log);
			if (iErr == 0) cStmap.nFlag[i] = 2;
		}
		// 被リンク数 0
		int n = 0;
		for (i = 0; i < cStmap.nLinks.length; i++) {
			if (cStmap.nLinks[i] == 0) {
				log.add('no link:' + cStmap.sLoc[i]);
				n += 1;
			}
		}
		log.add('no links:$n');

		// sitemapの書き出し
		C_writeMap cWrite = new C_writeMap();
		iErr = cWrite.write(cIni, cStmap);
		if (iErr != 0) print('Write Map Error:$iErr');

		// log書き出し
		iErr = log.write();
		return iErr;
	}

	void _debugprint(int i, String s0, String s1) {
		print('Count $i');
		print('FileName:$s0');
		print('Location:$s1');
	}
}